@@ -23,6 +23,10 @@ try:
     from urllib.parse import urlparse
 except ImportError:
     from urlparse import urlparse
+try:
+    from pathlib import Path
+except ImportError:
+    from pathlib2 import Path
 
 from distutils.spawn import find_executable
 from contextlib import contextmanager
@@ -32,7 +36,7 @@ from piptools.scripts.compile import get_pip_command
 from piptools import logging
 from piptools.exceptions import NoCandidateFound
 from pip.exceptions import DistributionNotFound
-from requests.exceptions import HTTPError
+from requests.exceptions import HTTPError, ConnectionError
 
 from .pep508checker import lookup
 from .environments import SESSION_IS_INTERACTIVE, PIPENV_MAX_ROUNDS, PIPENV_CACHE_DIR
@@ -41,7 +45,7 @@ specifiers = [k for k in lookup.keys()]
 
 # List of version control systems we support.
 VCS_LIST = ('git', 'svn', 'hg', 'bzr')
-SCHEME_LIST = ('http://', 'https://', 'ftp://', 'file:///')
+SCHEME_LIST = ('http://', 'https://', 'ftp://', 'file://')
 
 requests = requests.Session()
 
@@ -266,6 +270,58 @@ packages = [
 ]
 
 
+def get_requirement(dep):
+    """Pre-clean requirement strings passed to the requirements parser.
+
+    Ensures that we can accept both local and relative paths, file and VCS URIs,
+    remote URIs, and package names, and that we pass only valid requirement strings
+    to the requirements parser. Performs necessary modifications to requirements
+    object if the user input was a local relative path.
+    
+    :param str dep: A requirement line
+    :returns: :class:`requirements.Requirement` object
+    """
+    path = None
+    # Split out markers if they are present - similar to how pip does it
+    # See pip.req.req_install.InstallRequirement.from_line
+    if not any(dep.startswith(uri_prefix) for uri_prefix in SCHEME_LIST):
+        marker_sep = ';'
+    else:
+        marker_sep = '; '
+    if marker_sep in dep:
+        dep, markers = dep.split(marker_sep, 1)
+        markers = markers.strip()
+        if not markers:
+            markers = None
+    else:
+        markers = None
+    # Strip extras from the requirement so we can make a properly parseable req
+    dep, extras = pip.req.req_install._strip_extras(dep)
+    # Only operate on local, existing, non-URI formatted paths
+    if (is_file(dep) and isinstance(dep, six.string_types) and
+            not any(dep.startswith(uri_prefix) for uri_prefix in SCHEME_LIST)):
+        dep_path = Path(dep)
+        # Only parse if it is a file or an installable dir
+        if dep_path.is_file() or (dep_path.is_dir() and pip.utils.is_installable_dir(dep)):
+            if dep_path.is_absolute() or dep_path.as_posix() == '.':
+                path = dep_path.as_posix()
+            else:
+                path = get_converted_relative_path(dep)
+            dep = dep_path.resolve().as_uri()
+    req = [r for r in requirements.parse(dep)][0]
+    # If the result is a local file with a URI and we have a local path, unset the URI
+    # and set the path instead
+    if req.local_file and req.uri and not req.path and path:
+        req.path = path
+        req.uri = None
+    if markers:
+        req.markers = markers
+    if extras:
+        # Bizarrely this is also what pip does...
+        req.extras = [r for r in requirements.parse('fakepkg{0}'.format(extras))][0].extras
+    return req
+
+
 def cleanup_toml(tml):
     toml = tml.split('\n')
     new_toml = []
@@ -397,6 +453,78 @@ def prepare_pip_source_args(sources, pip_args=None):
     return pip_args
 
 
+def actually_resolve_reps(deps, index_lookup, markers_lookup, project, sources, verbose, clear, pre):
+
+    class PipCommand(pip.basecommand.Command):
+        """Needed for pip-tools."""
+        name = 'PipCommand'
+
+    constraints = []
+
+    for dep in deps:
+        t = tempfile.mkstemp(prefix='pipenv-', suffix='-requirement.txt')[1]
+        with open(t, 'w') as f:
+            f.write(dep)
+
+        if dep.startswith('-e '):
+            constraint = pip.req.InstallRequirement.from_editable(dep[len('-e '):])
+        else:
+            constraint = [c for c in pip.req.parse_requirements(t, session=pip._vendor.requests)][0]
+            # extra_constraints = []
+
+        if ' -i ' in dep:
+            index_lookup[constraint.name] = project.get_source(url=dep.split(' -i ')[1]).get('name')
+
+        if constraint.markers:
+            markers_lookup[constraint.name] = str(constraint.markers).replace('"', "'")
+
+        constraints.append(constraint)
+
+    pip_command = get_pip_command()
+
+    pip_args = []
+
+    if sources:
+        pip_args = prepare_pip_source_args(sources, pip_args)
+
+    if verbose:
+        print('Using pip: {0}'.format(' '.join(pip_args)))
+
+    pip_options, _ = pip_command.parse_args(pip_args)
+
+    session = pip_command._build_session(pip_options)
+    pypi = PyPIRepository(pip_options=pip_options, session=session)
+
+    if verbose:
+        logging.log.verbose = True
+
+
+    resolved_tree = set()
+
+    resolver = Resolver(constraints=constraints, repository=pypi, clear_caches=clear, prereleases=pre)
+    # pre-resolve instead of iterating to avoid asking pypi for hashes of editable packages
+    try:
+        resolved_tree.update(resolver.resolve(max_rounds=PIPENV_MAX_ROUNDS))
+    except (NoCandidateFound, DistributionNotFound, HTTPError) as e:
+        click.echo(
+            '{0}: Your dependencies could not be resolved. You likely have a mismatch in your sub-dependencies.\n  '
+            'You can use {1} to bypass this mechanism, then run {2} to inspect the situation.'
+            ''.format(
+                crayons.red('Warning', bold=True),
+                crayons.red('$ pipenv install --skip-lock'),
+                crayons.red('$ pipenv graph')
+            ),
+            err=True)
+
+        click.echo(crayons.blue(e))
+
+        if 'no version found at all' in str(e):
+            click.echo(crayons.blue('Please check your version specifier and version number. See PEP440 for more information.'))
+
+        raise RuntimeError
+
+    return resolved_tree
+
 def resolve_deps(deps, which, which_pip, project, sources=None, verbose=False, python=False, clear=False, pre=False):
     """Given a list of dependencies, return a resolved list of dependencies,
     using pip-tools -- and their hashes, using the warehouse API / pip.
@@ -406,76 +534,31 @@ def resolve_deps(deps, which, which_pip, project, sources=None, verbose=False, p
     markers_lookup = {}
 
     python_path = which('python')
+    backup_python_path = shellquote(sys.executable)
 
+    results = []
+
+    # First (proper) attempt:
     with HackedPythonVersion(python_version=python, python_path=python_path):
 
-        class PipCommand(pip.basecommand.Command):
-            """Needed for pip-tools."""
-            name = 'PipCommand'
-
-        constraints = []
-
-        for dep in deps:
-            t = tempfile.mkstemp(prefix='pipenv-', suffix='-requirement.txt')[1]
-            with open(t, 'w') as f:
-                f.write(dep)
-
-            if dep.startswith('-e '):
-                constraint = pip.req.InstallRequirement.from_editable(dep[len('-e '):])
-            else:
-                constraint = [c for c in pip.req.parse_requirements(t, session=pip._vendor.requests)][0]
-                # extra_constraints = []
-
-            if ' -i ' in dep:
-                index_lookup[constraint.name] = project.get_source(url=dep.split(' -i ')[1]).get('name')
-
-            if constraint.markers:
-                markers_lookup[constraint.name] = str(constraint.markers).replace('"', "'")
-
-            constraints.append(constraint)
-
-        pip_command = get_pip_command()
-
-        pip_args = []
-
-        if sources:
-            pip_args = prepare_pip_source_args(sources, pip_args)
-
-        if verbose:
-            print('Using pip: {0}'.format(' '.join(pip_args)))
-
-        pip_options, _ = pip_command.parse_args(pip_args)
-
-        session = pip_command._build_session(pip_options)
-        pypi = PyPIRepository(pip_options=pip_options, session=session)
-
-        if verbose:
-            logging.log.verbose = True
-
-        results = []
-        resolved_tree = set()
-
-        resolver = Resolver(constraints=constraints, repository=pypi, clear_caches=clear, prereleases=pre)
-        # pre-resolve instead of iterating to avoid asking pypi for hashes of editable packages
         try:
-            resolved_tree.update(resolver.resolve(max_rounds=PIPENV_MAX_ROUNDS))
-        except (NoCandidateFound, DistributionNotFound, HTTPError) as e:
-            click.echo(
-                '{0}: Your dependencies could not be resolved. You likely have a mismatch in your sub-dependencies.\n  '
-                'You can use {1} to bypass this mechanism, then run {2} to inspect the situation.'
-                ''.format(
-                    crayons.red('Warning', bold=True),
-                    crayons.red('$ pipenv install --skip-lock'),
-                    crayons.red('$ pipenv graph')
-                ),
-                err=True)
+            resolved_tree = actually_resolve_reps(deps, index_lookup, markers_lookup, project, sources, verbose, clear, pre)
+        except RuntimeError:
+            # Don't exit here, like usual.
+            resolved_tree = None
 
-            click.echo(crayons.blue(e))
+    # Second (last-resort) attempt:
+    if resolved_tree is None:
+        with HackedPythonVersion(python_version='.'.join([str(s) for s in sys.version_info[:3]]), python_path=backup_python_path):
+
+            try:
+                # Attempt to resolve again, with different Python version information,
+                # particularly for particularly particular packages.
+                resolved_tree = actually_resolve_reps(deps, index_lookup, markers_lookup, project, sources, verbose, clear, pre)
+            except RuntimeError:
+                sys.exit(1)
 
-            if 'no version found at all' in str(e):
-                click.echo(crayons.blue('Please check your version specifier and version number. See PEP440 for more information.'))
 
-            sys.exit(1)
 
     for result in resolved_tree:
         if not result.editable:
@@ -508,7 +591,7 @@ def resolve_deps(deps, which, which_pip, project, sources=None, verbose=False, p
                     if not collected_hashes:
                         collected_hashes = list(list(resolver.resolve_hashes([result]).items())[0][1])
 
-                except (ValueError, KeyError):
+                except (ValueError, KeyError, ConnectionError):
                     if verbose:
                         print('Error fetching {}'.format(name))
 
@@ -539,12 +622,11 @@ def convert_deps_from_pip(dep):
 
     dependency = {}
 
-    req = [r for r in requirements.parse(dep)][0]
+    req = get_requirement(dep)
     extras = {'extras': req.extras}
 
     # File installs.
-    if (req.uri or (os.path.isfile(req.path) if req.path else False) or
-            os.path.isfile(req.name)) and not req.vcs:
+    if (req.uri or req.path or (os.path.isfile(req.name) if req.name else False)) and not req.vcs:
         # Assign a package name to the file, last 7 of it's sha256 hex digest.
         if not req.uri and not req.path:
             req.path = os.path.abspath(req.name)
@@ -552,28 +634,26 @@ def convert_deps_from_pip(dep):
         hashable_path = req.uri if req.uri else req.path
         req.name = hashlib.sha256(hashable_path.encode('utf-8')).hexdigest()
         req.name = req.name[len(req.name) - 7:]
-
         # {path: uri} TOML (spec 4 I guess...)
         if req.uri:
             dependency[req.name] = {'file': hashable_path}
         else:
             dependency[req.name] = {'path': hashable_path}
 
+        if req.extras:
+            dependency[req.name].update(extras)
+
         # Add --editable if applicable
         if req.editable:
             dependency[req.name].update({'editable': True})
 
     # VCS Installs. Extra check for unparsed git over SSH
-    if req.vcs or is_vcs(req.path):
+    elif req.vcs or is_vcs(req.path):
         if req.name is None:
             raise ValueError('pipenv requires an #egg fragment for version controlled '
                              'dependencies. Please install remote dependency '
                              'in the form {0}#egg=<package-name>.'.format(req.uri))
 
-        # Extras: e.g. #egg=requests[security]
-        if req.extras:
-            dependency[req.name] = extras
-
         # Set up this requirement as a proper VCS requirement if it was not
         if not req.vcs and req.path.startswith(VCS_LIST):
             req.vcs = [vcs for vcs in VCS_LIST if req.path.startswith(vcs)][0]
@@ -581,7 +661,9 @@ def convert_deps_from_pip(dep):
             req.path = None
 
         # Crop off the git+, etc part.
-        dependency.setdefault(req.name, {}).update({req.vcs: req.uri[len(req.vcs) + 1:]})
+        if req.uri.startswith('{0}+'.format(req.vcs)):
+            req.uri = req.uri[len(req.vcs) + 1:]
+        dependency.setdefault(req.name, {}).update({req.vcs: req.uri})
 
         # Add --editable, if it's there.
         if req.editable:
@@ -595,6 +677,10 @@ def convert_deps_from_pip(dep):
         if req.revision:
             dependency[req.name].update({'ref': req.revision})
 
+        # Extras: e.g. #egg=requests[security]
+        if req.extras:
+            dependency[req.name].update({'extras': req.extras})
+
     elif req.extras or req.specs:
 
         specs = None
@@ -620,7 +706,6 @@ def convert_deps_from_pip(dep):
         for key in dependency.copy():
             if not hasattr(dependency[key], 'keys'):
                 del dependency[key]
-
     return dependency
 
 
@@ -771,10 +856,24 @@ def is_vcs(pipfile_entry):
     if hasattr(pipfile_entry, 'keys'):
         return any(key for key in pipfile_entry.keys() if key in VCS_LIST)
     elif isinstance(pipfile_entry, six.string_types):
-        return pipfile_entry.startswith(VCS_LIST)
+        # Add scheme for parsing purposes, this is also what pip does
+        if pipfile_entry.startswith('git+') and '://' not in pipfile_entry:
+            pipfile_entry = pipfile_entry.replace('git+', 'git+ssh://')
+        return bool(requirements.requirement.VCS_REGEX.match(pipfile_entry))
     return False
 
 
+def is_installable_file(path):
+    """Determine if a path can potentially be installed"""
+    if hasattr(path, 'keys') and any(key for key in path.keys() if key in ['file', 'path']):
+        path = urlparse(path['file']).path if 'file' in path else path['path']
+    if not isinstance(path, six.string_types) or path == '*':
+        return False
+    lookup_path = Path(path)
+    return lookup_path.is_file() or (lookup_path.is_dir() and
+            pip.utils.is_installable_dir(lookup_path.resolve().as_posix()))
+
+
 def is_file(package):
     """Determine if a package name is for a File dependency."""
     if hasattr(package, 'keys'):
@@ -820,21 +919,100 @@ def proper_case(package_name):
     return good_name
 
 
-def split_vcs(split_file):
-    """Split VCS dependencies out from file."""
+def split_section(input_file, section_suffix, test_function):
+    """
+    Split a pipfile or a lockfile section out by section name and test function
+    
+        :param dict input_file: A dictionary containing either a pipfile or lockfile
+        :param str section_suffix: A string of the name of the section
+        :param func test_function: A test function to test against the value in the key/value pair
+    
+    >>> split_section(my_lockfile, 'vcs', is_vcs)
+    {
+        'default': {
+            "six": {
+                "hashes": [
+                    "sha256:832dc0e10feb1aa2c68dcc57dbb658f1c7e65b9b61af69048abc87a2db00a0eb",
+                    "sha256:70e8a77beed4562e7f14fe23a786b54f6296e34344c23bc42f07b15018ff98e9"
+                ],
+                "version": "==1.11.0"
+            }
+        },
+        'default-vcs': {
+            "e1839a8": {
+                "editable": true,
+                "path": "."
+            }
+        }
+    }
+    """
+    pipfile_sections = ('packages', 'dev-packages')
+    lockfile_sections = ('default', 'develop')
+    if any(section in input_file for section in pipfile_sections):
+        sections = pipfile_sections
+    elif any(section in input_file for section in lockfile_sections):
+        sections = lockfile_sections
+    else:
+        # return the original file if we can't find any pipfile or lockfile sections
+        return input_file
 
-    if 'packages' in split_file or 'dev-packages' in split_file:
-        sections = ('packages', 'dev-packages')
-    elif 'default' in split_file or 'develop' in split_file:
-        sections = ('default', 'develop')
-
-    # For each vcs entry in a given section, move it to section-vcs.
     for section in sections:
-        entries = split_file.get(section, {})
-        vcs_dict = dict((k, entries.pop(k)) for k in list(entries.keys()) if is_vcs(entries[k]))
-        split_file[section + '-vcs'] = vcs_dict
+        split_dict = {}
+        entries = input_file.get(section, {})
+        for k in list(entries.keys()):
+            if test_function(entries.get(k)):
+                split_dict[k] = entries.pop(k)
+        input_file['-'.join([section, section_suffix])] = split_dict
+    return input_file
 
-    return split_file
+
+def split_file(file_dict):
+    """Split VCS and editable dependencies out from file."""
+    sections = {
+        'vcs': is_vcs,
+        'editable': lambda x: hasattr(x, 'keys') and x.get('editable')
+    }
+    for k, func in sections.items():
+        file_dict = split_section(file_dict, k, func)
+    return file_dict
+
+
+def merge_deps(file_dict, project, dev=False, requirements=False, ignore_hashes=False, blocking=False, only=False):
+    """
+    Given a file_dict, merges dependencies and converts them to pip dependency lists.
+        :param dict file_dict: The result of calling :func:`pipenv.utils.split_file`
+        :param :class:`pipenv.project.Project` project: Pipenv project
+        :param bool dev=False: Flag indicating whether dev dependencies are to be installed 
+        :param bool requirements=False: Flag indicating whether to use a requirements file
+        :param bool ignore_hashes=False:
+        :param bool blocking=False:
+        :param bool only=False:
+        :return: Pip-converted 3-tuples of [deps, requirements_deps]
+    """
+    deps = []
+    requirements_deps = []
+
+    for section in list(file_dict.keys()):
+        # Turn develop-vcs into ['develop', 'vcs']
+        section_name, suffix = section.rsplit('-', 1) if '-' in section and not section == 'dev-packages' else (section, None)
+        if not file_dict[section] or section_name not in ('dev-packages', 'packages', 'default', 'develop'):
+            continue
+        is_dev = section_name in ('dev-packages', 'develop')
+
+        if ignore_hashes:
+            for k, v in file_dict[section]:
+                if 'hash' in v:
+                    del v['hash']
+
+        # Block and ignore hashes for all suffixed sections (vcs/editable)
+        no_hashes = True if suffix else ignore_hashes
+        block = True if suffix else blocking
+        include_index = True if not suffix else False
+        converted = convert_deps_to_pip(file_dict[section], project, r=False, include_index=include_index)
+        deps.extend((d, no_hashes, block) for d in converted)
+        if dev and is_dev and requirements:
+            requirements_deps.extend((d, no_hashes, block) for d in converted)
+    return deps, requirements_deps
 
 
 def recase_file(file_dict):
@@ -882,6 +1060,11 @@ def find_windows_executable(bin_path, exe_name):
     return find_executable(exe_name)
 
 
+def get_converted_relative_path(path, relative_to=os.curdir):
+    """Given a vague relative path, return the path relative to the given location"""
+    return os.path.join('.', os.path.relpath(path, start=relative_to))
+
+
 def walk_up(bottom):
     """Mimic os.walk, but walk 'up' instead of down the directory tree.
     From: https://gist.github.com/zdavkeos/1098474
@@ -975,7 +1158,7 @@ def touch_update_stamp():
     mkdir_p(PIPENV_CACHE_DIR)
     p = os.sep.join((PIPENV_CACHE_DIR, '.pipenv_update_check'))
     try:
-        os.utime(p)
-    except FileNotFoundError:
+        os.utime(p, None)
+    except OSError:
         with open(p, 'w') as fh:
             fh.write('')
